from __future__ import annotations

from enum import Enum
from datetime import datetime
from pathlib import Path
import re
import sys
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    ForwardRef,
    Iterable,
    Literal,
    Match,
    TypeVar,
    Union,
    cast,
    final,
    overload, Generic,
)

from tarina import DateParser, Empty, lang

from .core import BasePattern, MatchMode, ResultFlag, ValidateResult
from .exception import MatchFailed
from .util import TPattern

TOrigin = TypeVar("TOrigin")
TDefault = TypeVar("TDefault")
_T = TypeVar("_T")
_T1 = TypeVar("_T1")


class DirectPattern(BasePattern[TOrigin, TOrigin]):
    """直接判断"""

    __slots__ = ("target",)

    def __init__(self, target: TOrigin, alias: str | None = None):
        self.target = target
        super().__init__(mode=MatchMode.KEEP, origin=type(target), alias=alias or str(target))

    def match(self, input_: Any):
        if input_ != self.target:
            raise MatchFailed(
                lang.require("nepattern", "content_error").format(target=input_, expected=self.target)
            )
        return input_

    @overload
    def validate(self, input_: TOrigin) -> ValidateResult[TOrigin, Literal[ResultFlag.VALID]]:
        ...

    @overload
    def validate(self, input_: _T) -> ValidateResult[_T, Literal[ResultFlag.ERROR]]:
        ...

    @overload
    def validate(self, input_: TOrigin, default: Any) -> ValidateResult[TOrigin, Literal[ResultFlag.VALID]]:
        ...

    @overload
    def validate(
        self, input_: Any, default: TDefault
    ) -> ValidateResult[TDefault, Literal[ResultFlag.DEFAULT]]:
        ...

    def validate(self, input_: Any, default: Union[TDefault, Empty] = Empty) -> ValidateResult[TOrigin | TDefault, ResultFlag]:  # type: ignore
        if input_ == self.target:
            return ValidateResult(input_, flag=ResultFlag.VALID)
        e = MatchFailed(
            lang.require("nepattern", "content_error").format(target=input_, expected=self.target)
        )
        if default is Empty:
            return ValidateResult(error=e, flag=ResultFlag.ERROR)
        return ValidateResult(default, flag=ResultFlag.DEFAULT)  # type: ignore

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, DirectPattern) and self.target == other.target

class DirectTypePattern(BasePattern[TOrigin, TOrigin]):
    """直接类型判断"""

    __slots__ = ("target",)

    def __init__(self, target: type[TOrigin], alias: str | None = None):
        self.target = target
        super().__init__(mode=MatchMode.KEEP, origin=target, alias=alias or target.__name__)

    def match(self, input_: Any):
        if not isinstance(input_, self.target):
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected=self.target
                )
            )
        return input_

    @overload
    def validate(self, input_: TOrigin) -> ValidateResult[TOrigin, Literal[ResultFlag.VALID]]:
        ...

    @overload
    def validate(self, input_: _T) -> ValidateResult[_T, Literal[ResultFlag.ERROR]]:
        ...

    @overload
    def validate(self, input_: TOrigin, default: Any) -> ValidateResult[TOrigin, Literal[ResultFlag.VALID]]:
        ...

    @overload
    def validate(
        self, input_: Any, default: TDefault
    ) -> ValidateResult[TDefault, Literal[ResultFlag.DEFAULT]]:
        ...

    def validate(self, input_: Any, default: Union[TDefault, Empty] = Empty) -> ValidateResult[TOrigin | TDefault, ResultFlag]:  # type: ignore
        if isinstance(input_, self.target):
            return ValidateResult(input_, flag=ResultFlag.VALID)
        e = MatchFailed(
            lang.require("nepattern", "type_error").format(
                type=input_.__class__, target=input_, expected=self.target
            )
        )
        if default is Empty:
            return ValidateResult(error=e, flag=ResultFlag.ERROR)
        return ValidateResult(default, flag=ResultFlag.DEFAULT)  # type: ignore

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, DirectTypePattern) and self.target == other.target

class RegexPattern(BasePattern[Match[str], str]):
    """针对正则的特化匹配，支持正则组"""

    def __init__(self, pattern: str | TPattern, alias: str | None = None):
        super().__init__("", origin=Match[str], alias=alias or "regex[:group]")  # type: ignore
        self.regex_pattern = re.compile(pattern)
        self.pattern = self.regex_pattern.pattern

    def match(self, input_: Any) -> Match[str]:
        if not isinstance(input_, str):
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected="str"
                )
            )
        if mat := self.regex_pattern.match(input_):
            return mat
        raise MatchFailed(
            lang.require("nepattern", "content_error").format(target=input_, expected=self.pattern)
        )

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, RegexPattern) and self.pattern == other.pattern


class UnionPattern(BasePattern[Any, _T]):
    """多类型参数的匹配"""

    optional: bool
    for_validate: list[BasePattern]
    for_equal: list[str | object]

    __slots__ = ("base", "optional", "for_validate", "for_equal")

    def __init__(self, base: Iterable[BasePattern[Any, _T] | _T]):
        self.base = list(base)
        self.optional = False
        self.for_validate = []
        self.for_equal = []

        for arg in self.base:
            if arg == NONE:
                self.optional = True
                self.for_equal.append(None)
            elif isinstance(arg, BasePattern):
                self.for_validate.append(arg)
            else:
                self.for_equal.append(arg)
        alias_content = "|".join([repr(a) for a in self.for_validate] + [repr(a) for a in self.for_equal])
        super().__init__(mode=MatchMode.KEEP, origin=str, alias=alias_content)

    def match(self, text: Any):
        if not text:
            text = None
        if text not in self.for_equal:
            for pat in self.for_validate:
                if (res := pat.validate(text)).success:
                    return res.value()
            raise MatchFailed(
                lang.require("nepattern", "content_error").format(target=text, expected=self.alias)
            )
        return text
    
    @classmethod
    def _(cls, *types: type[_T1]) -> UnionPattern[_T1]:
        from .main import parser

        return cls([parser(i) for i in types])  # type: ignore

    def __calc_repr__(self):
        return "|".join(repr(a) for a in (*self.for_validate, *self.for_equal))

    def __or__(self, other: BasePattern[Any, _T1]) -> UnionPattern[Union[_T, _T1]]:
        return UnionPattern([*self.base, other])  # type: ignore

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, UnionPattern) and self.base == other.base

TSeq = TypeVar("TSeq", list, tuple, set)

class IterMode(str, Enum):
    PRE = "pre"
    SUF = "suf"
    ALL = "all"

TIterMode = TypeVar("TIterMode", bound=IterMode)

class SequencePattern(BasePattern[TSeq, Union[str, TSeq]], Generic[TSeq, TIterMode]):
    """匹配列表或者元组或者集合"""

    base: BasePattern
    itermode: TIterMode

    def __init__(self, form: type[TSeq], base: BasePattern, mode: TIterMode = IterMode.ALL):
        self.base = base
        self.itermode = mode
        if form is list:
            super().__init__(r"\[(.+?)\]", MatchMode.REGEX_CONVERT, form, alias=f"list[{base}]")
        elif form is tuple:
            super().__init__(r"\((.+?)\)", MatchMode.REGEX_CONVERT, form, alias=f"tuple[{base}]")
        elif form is set:
            super().__init__(r"\{(.+?)\}", MatchMode.REGEX_CONVERT, form, alias=f"set[{base}]")
        else:
            raise ValueError(lang.require("nepattern", "sequence_form_error").format(target=str(form)))

    def match(self, text: Any):
        _res = self._MATCHES[MatchMode.REGEX_CONVERT](self, text)  # type: ignore
        _max = 0
        success: list[tuple[int, Any]] = []
        fail: list[tuple[int, MatchFailed]] = []
        for _max, s in enumerate(re.split(r"\s*,\s*", _res) if isinstance(_res, str) else _res):
            try:
                success.append((_max, self.base.match(s)))
            except MatchFailed:
                fail.append((_max, MatchFailed(f"{s} is not matched with {self.base}")))

        if (
            (self.itermode == IterMode.ALL and fail)
            or (self.itermode == IterMode.PRE and fail and fail[0][0] == 0)
            or (self.itermode == IterMode.SUF and fail and fail[-1][0] == _max)
        ):
            raise fail[0][1]
        if self.itermode == IterMode.PRE and fail:
            return self.origin(i[1] for i in success if i[0] < fail[0][0])
        if self.itermode == IterMode.SUF and fail:
            return self.origin(i[1] for i in success if i[0] > fail[-1][0])
        return self.origin(i[1] for i in success)

    def __calc_repr__(self):
        return f"{self.origin.__name__}[{self.base}]"


TKey = TypeVar("TKey")
TVal = TypeVar("TVal")


class MappingPattern(
    BasePattern[Dict[TKey, TVal], Union[str, Dict[TKey, TVal]]],
    Generic[TKey, TVal, TIterMode],
):
    """匹配字典或者映射表"""

    key: BasePattern[TKey, Any]
    value: BasePattern[TVal, Any]
    itermode: TIterMode


    def __init__(
        self, 
        arg_key: BasePattern[TKey, Any], 
        arg_value: BasePattern[TVal, Any],
        mode: TIterMode = IterMode.ALL
    ):
        self.key = arg_key
        self.value = arg_value
        self.itermode = mode
        super().__init__(
            r"\{(.+?)\}",
            MatchMode.REGEX_CONVERT,
            dict,
            alias=f"dict[{self.key}, {self.value}]",
        )
        self.converter = lambda _, x: x[1]

    def match(self, input_: str | dict):
        _res = self._MATCHES[MatchMode.REGEX_CONVERT](self, input_)  # type: ignore
        success: list[tuple[int, Any, Any]] = []
        fail: list[tuple[int, MatchFailed]] = []
        _max = 0

        def _generator_items(res: str | dict):
            if isinstance(res, dict):
                yield from res.items()
                return
            for m in re.split(r"\s*,\s*", res):
                yield re.split(r"\s*[:=]\s*", m)

        for _max, item in enumerate(_generator_items(_res)):
            k, v = item
            try:
                success.append((_max, self.key.match(k), self.value.match(v)))
            except MatchFailed:
                fail.append(
                    (
                        _max,
                        MatchFailed(f"{k}: {v} is not matched with {self.key}: {self.value}"),
                    )
                )
        if (
            (self.itermode == IterMode.ALL and fail)
            or (self.itermode == IterMode.PRE and fail and fail[0][0] == 0)
            or (self.itermode == IterMode.SUF and fail and fail[-1][0] == _max)
        ):
            raise fail[0][1]
        if self.itermode == IterMode.PRE and fail:
            return {i[1]: i[2] for i in success if i[0] < fail[0][0]}
        if self.itermode == IterMode.SUF and fail:
            return {i[1]: i[2] for i in success if i[0] > fail[-1][0]}
        return {i[1]: i[2] for i in success}

    def __calc_repr__(self):
        return f"dict[{self.key.origin.__name__}, {self.value}]"


_TCase = TypeVar("_TCase")
_TSwtich = TypeVar("_TSwtich")


class SwitchPattern(BasePattern[_TCase, _TSwtich]):
    switch: dict[_TSwtich | ellipsis, _TCase]

    __slots__ = ("switch",)

    def __init__(self, data: dict[_TSwtich | ellipsis, _TCase]):
        self.switch = data
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=type(list(data.values())[0]))

    def __calc_repr__(self):
        return "|".join(f"{k}" for k in self.switch if k != Ellipsis)

    def match(self, input_: Any) -> _TCase:
        try:
            return self.switch[input_]
        except KeyError as e:
            if Ellipsis in self.switch:
                return self.switch[...]
            raise MatchFailed(
                lang.require("nepattern", "content_error").format(target=input_, expected=self._repr)
            ) from e

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, SwitchPattern) and self.switch == other.switch


class ForwardRefPattern(BasePattern[Any, Any]):
    def __init__(self, ref: ForwardRef):
        self.ref = ref
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=Any, alias=ref.__forward_arg__)

    def match(self, input_: Any):
        if isinstance(input_, str) and input_ == self.ref.__forward_arg__:
            return input_
        _main = sys.modules["__main__"]
        if sys.version_info < (3, 9):  # pragma: no cover
            origin = self.ref._evaluate(_main.__dict__, _main.__dict__)
        else:  # pragma: no cover
            origin = self.ref._evaluate(_main.__dict__, _main.__dict__, frozenset())  # type: ignore
        if not isinstance(input_, origin):  # type: ignore
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected=self.ref.__forward_arg__
                )
            )
        return input_

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, ForwardRefPattern) and self.ref == other.ref


class AntiPattern(BasePattern[TOrigin, Any]):
    def __init__(self, pattern: BasePattern[TOrigin, Any]):
        self.base: BasePattern[TOrigin, Any] = pattern
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=pattern.origin, alias=f"!{pattern}")

    @overload
    def validate(self, input_: TOrigin) -> ValidateResult[Any, Literal[ResultFlag.ERROR]]:
        ...

    @overload
    def validate(self, input_: _T) -> ValidateResult[_T, Literal[ResultFlag.VALID]]:
        ...

    @overload
    def validate(
        self, input_: TOrigin, default: TDefault
    ) -> ValidateResult[TDefault, Literal[ResultFlag.DEFAULT]]:
        ...

    @overload
    def validate(self, input_: _T, default: Any) -> ValidateResult[_T, Literal[ResultFlag.VALID]]:
        ...

    def validate(self, input_: _T, default: Union[TDefault, Empty] = Empty) -> ValidateResult[_T | TDefault, ResultFlag]:  # type: ignore
        """
        对传入的值进行反向验证，返回可能的匹配与转化结果。

        若传入默认值，验证失败会返回默认值
        """
        try:
            res = self.base.match(input_)
        except MatchFailed:
            return ValidateResult(value=input_, flag=ResultFlag.VALID)
        else:  # pragma: no cover
            for i in self.base.validators + self.validators:
                if not i(res):
                    return ValidateResult(value=input_, flag=ResultFlag.VALID)
            if default is Empty:
                return ValidateResult(
                    error=MatchFailed(
                        lang.require("nepattern", "content_error").format(target=input_, expected=self._repr)
                    ),
                    flag=ResultFlag.ERROR,
                )
            if TYPE_CHECKING:
                default = cast(TDefault, default)
            return ValidateResult(default, flag=ResultFlag.DEFAULT)

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, AntiPattern) and self.base == other.base

TInput = TypeVar("TInput")


class CustomMatchPattern(BasePattern[TOrigin, TInput]):
    def __init__(
        self,
        origin: type[TOrigin],
        func: Callable[[BasePattern, TInput], TOrigin | None],
        alias: str | None = None,
    ):
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=origin, alias=alias or func.__name__)
        self.__func__ = func
        self.match = func.__get__(self)  # type: ignore

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, CustomMatchPattern) and self.__func__ == other.__func__

NONE = BasePattern(mode=MatchMode.KEEP, origin=None, alias="none")  # type: ignore

ANY = BasePattern(mode=MatchMode.KEEP, origin=Any, alias="any")
"""匹配任意内容的表达式"""

def _any_str(_, x: Any) -> str:
    return str(x)

AnyString = CustomMatchPattern(str, _any_str, "any_str")
"""匹配任意内容并转为字符串的表达式"""

@final
class StrPattern(BasePattern[str, str]):
    def __init__(self):
        super().__init__(mode=MatchMode.KEEP, origin=str, alias="str", accepts=str)
    
    def match(self, input_: str) -> str:
        if not isinstance(input_, str):  # pragma: no cover
            raise MatchFailed(
                lang.require("nepattern", "type_error")
                .format(type=input_.__class__, target=input_, expected="str")
            )
        return input_
    
    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, StrPattern)
    
STRING = StrPattern()

@final
class IntPattern(BasePattern[int, Union[str, int]]):
    def __init__(self):
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=int, alias="int")

    def match(self, input_: Union[str, int]) -> int:
        if isinstance(input_, int):
            return input_
        if not isinstance(input_, str):
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected="int | str"
                )
            )
        try:
            return int(input_)
        except ValueError as e:
            raise MatchFailed(
                lang.require("nepattern", "content_error").format(target=input_, expected="int")
            ) from e

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, IntPattern)


INTEGER = IntPattern()
"""整形数表达式，只能接受整数样式的量"""


@final
class FloatPattern(BasePattern[float, Union[str, int, float]]):
    def __init__(self):
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=float, alias="float")

    def match(self, input_: Union[str, float, int]) -> float:
        if isinstance(input_, float):
            return input_
        if not isinstance(input_, (str, int)):
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected="str | int | float"
                )
            )
        try:
            return float(input_)
        except ValueError as e:
            raise MatchFailed(
                lang.require("nepattern", "content_error").format(target=input_, expected="float")
            ) from e

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, FloatPattern)

FLOAT = FloatPattern()
"""浮点数表达式"""


@final
class NumberPattern(BasePattern[Union[int, float], Union[str, int, float]]):
    def __init__(self):
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=Union[int, float], alias="number")

    def match(self, input_: Union[str, float]) -> float:
        if isinstance(input_, (float, int)):
            return input_
        if not isinstance(input_, str):
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected="str | int | float"
                )
            )
        try:
            res = float(input_)
            return int(res) if res.is_integer() else res
        except ValueError as e:
            raise MatchFailed(
                lang.require("nepattern", "content_error").format(target=input_, expected="int | float")
            ) from e

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, NumberPattern)

NUMBER = NumberPattern()
"""一般数表达式，既可以浮点数也可以整数 """


@final
class BoolPattern(BasePattern[bool, Union[str, bool]]):
    def __init__(self):
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=bool, alias="bool")

    _BOOL = {"true": True, "false": False, "True": True, "False": False}

    def match(self, input_: Union[str, bool]) -> bool:
        if isinstance(input_, bool):
            return input_
        if not isinstance(input_, str):
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected="str | bool"
                )
            )
        if input_ in self._BOOL:
            return self._BOOL[input_]
        raise MatchFailed(lang.require("nepattern", "content_error").format(target=input_, expected="bool"))
    
    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, BoolPattern)


BOOLEAN = BoolPattern()
"""布尔表达式，只能接受true或false样式的量"""

LIST = BasePattern(r"(\[.+?\])", MatchMode.REGEX_CONVERT, list, alias="list")
TUPLE = BasePattern(r"(\(.+?\))", MatchMode.REGEX_CONVERT, tuple, alias="tuple")
SET = BasePattern(r"(\{.+?\})", MatchMode.REGEX_CONVERT, set, alias="set")
DICT = BasePattern(r"(\{.+?\})", MatchMode.REGEX_CONVERT, dict, alias="dict")

EMAIL = BasePattern(r"(?:[\w\.+-]+)@(?:[\w\.-]+)\.(?:[\w\.-]+)", MatchMode.REGEX_MATCH, alias="email")
"""匹配邮箱地址的表达式"""

IP = BasePattern(
    r"(?:(?:[01]{0,1}\d{0,1}\d|2[0-4]\d|25[0-5])\.){3}(?:[01]{0,1}\d{0,1}\d|2[0-4]\d|25[0-5]):?(?:\d+)?",
    MatchMode.REGEX_MATCH,
    alias="ip",
)
"""匹配Ip地址的表达式"""

URL = BasePattern(
    r"(?:\w+://)?[a-zA-Z0-9][-a-zA-Z0-9]{0,62}(?:\.[a-zA-Z0-9][-a-zA-Z0-9]{0,62})+(?::[0-9]{1,5})?[-a-zA-Z0-9()@:%_\\\+\.~#?&//=]*",
    MatchMode.REGEX_MATCH,
    alias="url",
)
"""匹配网页链接的表达式"""


@final
class HexPattern(BasePattern[int, str]):
    def __init__(self):
        super().__init__(mode=MatchMode.TYPE_CONVERT, origin=int, alias="hex", accepts=str)

    def match(self, input_: str) -> int:
        if not isinstance(input_, str):
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected="str"
                )
            )
        try:
            return int(input_, 16)
        except ValueError as e:
            raise MatchFailed(
                lang.require("nepattern", "content_error").format(target=input_, expected="hex")
            ) from e
    
    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, HexPattern)

HEX = HexPattern()
"""匹配16进制数的表达式"""

HEX_COLOR = BasePattern(r"(#[0-9a-fA-F]{6})", MatchMode.REGEX_CONVERT, str, lambda _, x: x[1][1:], "color")
"""匹配16进制颜色代码的表达式"""


@final
class DateTimePattern(BasePattern[datetime, Union[str, int, float]]):
    def __init__(self):
        super().__init__(
            mode=MatchMode.TYPE_CONVERT, origin=datetime, alias="datetime", accepts=Union[str, int, float]
        )

    def match(self, input_: Union[str, int, float]) -> datetime:
        if isinstance(input_, (int, float)):
            return datetime.fromtimestamp(input_)
        if not isinstance(input_, str):
            raise MatchFailed(
                lang.require("nepattern", "type_error").format(
                    type=input_.__class__, target=input_, expected="str | int | float"
                )
            )
        return DateParser.parse(input_)

    def __calc_eq__(self, other):  # pragma: no cover
        return isinstance(other, DateTimePattern)

DATETIME = DateTimePattern()
"""匹配时间的表达式"""


StrPath = BasePattern(mode=MatchMode.TYPE_CONVERT, origin=Path, alias="path", accepts=str)

PathFile = BasePattern(
    mode=MatchMode.TYPE_CONVERT,
    origin=bytes,
    alias="file",
    accepts=Path,
    previous=StrPath,
    converter=lambda _, x: x.read_bytes() if x.exists() and x.is_file() else None,
)
